import os

import fastapi
import pytest
import yaml
from fastapi import testclient
from pydantic import dataclasses as pydantic_dataclasses

import fixieai
from fixieai import agents
from fixieai.agents import agent_base

agent_id = "dummy"


@pytest.fixture(autouse=True)
def mock_token_verifier(mocker):
    return mocker.patch.object(
        agent_base.VerifiedTokenClaims,
        "from_token",
        return_value=agent_base.VerifiedTokenClaims(agent_id="fake agent id"),
    )


@pytest.fixture
def dummy_agent(mocker):
    mocker.patch.dict(os.environ, {"FIXIE_ALLOWED_AGENT_ID": "fake agent id"})

    @pydantic_dataclasses.dataclass
    class DummyMetadata:
        dummy_metadata: bool = True

    class DummyAgent(agents.AgentBase):
        def metadata(self):
            return DummyMetadata()

        def validate(self):
            pass

    agent = DummyAgent(
        oauth_params=fixieai.OAuthParams(
            client_id="dummy",
            client_secret="dummy",
            auth_uri="dummy",
            token_uri="dummy",
            scopes=["dummy"],
        ),
    )

    @agent.register_func
    def simple1(query: agents.Message) -> str:
        return "Simple response 1"

    @agent.register_func()
    def simple2(query: agents.Message):
        return "Simple response 2"

    @agent.register_func(func_name="custom")
    def simple3(query):
        return "Simple response custom"

    return agent


def test_simple_agent_handshake(dummy_agent, mocker):
    fast_api = fastapi.FastAPI()
    fast_api.include_router(dummy_agent.api_router())

    client = testclient.TestClient(fast_api)

    # Use a small chunk size to exercise chunking logic
    mocker.patch.object(agent_base, "_RESPONSE_CHUNK_SIZE", 2)
    response = client.get("/", headers={"Authorization": "Bearer fixie-test-token"})
    assert response.status_code == 200
    assert response.headers["content-type"] == "application/yaml"
    yaml_content = yaml.load(response.content, Loader=yaml.Loader)
    assert yaml_content == {"dummy_metadata": True}


def test_simple_agent_func_calls(dummy_agent, mock_token_verifier):
    fast_api = fastapi.FastAPI()
    fast_api.include_router(dummy_agent.api_router())

    client = testclient.TestClient(fast_api, raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}

    # Test Func[simple1]
    response = client.post(
        "/simple1", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 200
    json = response.json()
    assert json == {"message": {"text": "Simple response 1", "embeds": {}}}
    mock_token_verifier.assert_called_once_with(
        "fixie-test-token", dummy_agent._jwks_client, dummy_agent._allowed_agent_id
    )

    # Test Func[simple2]
    response = client.post(
        "/simple2", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 200
    json = response.json()
    assert json == {"message": {"text": "Simple response 2", "embeds": {}}}

    # Test Func[custom]
    response = client.post(
        "/custom", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 200
    json = response.json()
    assert json == {"message": {"text": "Simple response custom", "embeds": {}}}

    # Test non-existing Func[] returns 404: Not Found
    response = client.post(
        "/simple3", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 404

    # Test Func[simple1] with bad arguments returns 422: Unprocessable Entity
    response = client.post(
        "/simple1", json={"message": {"ttt": "Howdy"}}, headers=headers
    )
    assert response.status_code == 422

    # Test Func[__init__] 404: Not Found
    response = client.post(
        "/__init__", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 404

    # Test Func without auth header returns 401: Unauthorized
    response = client.post("/simple1", json={"message": {"text": "Howdy"}})
    assert response.status_code == 403


def test_registering_func_with_custom_name(dummy_agent):
    dummy_agent.register_func(lambda query: "1", func_name="name1")
    dummy_agent.register_func(lambda query: "2", func_name="name2")
    dummy_agent.register_func(lambda query: "3", func_name="name3")

    def invoke_func(name: str):
        responses = dummy_agent._funcs[name](
            fixieai.AgentQuery(fixieai.Message("test")), "agent-id"
        )
        return next(iter(responses)).message.text

    assert invoke_func("name1") == "1"
    assert invoke_func("name2") == "2"
    assert invoke_func("name3") == "3"


def test_invalid_token(dummy_agent, mock_token_verifier):
    mock_token_verifier.return_value = None

    fast_api = fastapi.FastAPI()
    fast_api.include_router(dummy_agent.api_router())
    client = testclient.TestClient(fast_api, raise_server_exceptions=False)
    headers = {"Authorization": "Bearer fixie-test-token"}
    response = client.post(
        "/simple1", json={"message": {"text": "Howdy"}}, headers=headers
    )
    assert response.status_code == 403



from __future__ import annotations

import collections.abc
import inspect
import typing
from typing import (
    Any,
    Callable,
    Iterable,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    get_type_hints,
)

from fixieai.agents import api
from fixieai.agents import oauth
from fixieai.agents import user_storage

# An ArgumentMapper converts a query+agent ID into a supported func argument (e.g. "query" or "user_storage")
ArgumentMapper = Callable[[api.AgentQuery, str], Any]

# BoundArgumentMapper is an ArgumentMapper that has been bound to a named parameter.
BoundArgumentMapper = Tuple[str, ArgumentMapper]


class AgentFunc:
    """A Python function that can be invoked by Fixie.

    Wrapped functions can take up to three arguments:

    1. A message or query of type: `str`, `api.AgentMessage`, or `api.AgentQuery`;
       name: "query", or "message"; or the first parameter if no other rules apply.
    2. A user storage object of type `user_storage.UserStorage` or name "user_storage".
    3. An OAuth handler of type `oauth.OAuthHandler` or name "oauth_handler".

    The semantics of each parameter are inferred by type annotation or name/position if no
    type annotations are present.
    """

    def __init__(
        self,
        impl: Callable,
        argument_mappers: Iterable[BoundArgumentMapper],
        allow_multiple_responses: bool,
    ):
        self._impl = impl
        self._argument_mappers = tuple(argument_mappers)
        self._allow_multiple_responses = allow_multiple_responses

    def __call__(
        self, agent_query: api.AgentQuery, agent_id: str
    ) -> Iterable[api.AgentResponse]:
        kwargs = {}
        for name, mapper in self._argument_mappers:
            kwargs[name] = mapper(agent_query, agent_id)

        result = self._impl(**kwargs)
        return _adapt_func_result(result, self._allow_multiple_responses)

    @classmethod
    def create(
        cls,
        func: Callable,
        oauth_params: Optional[oauth.OAuthParams],
        default_message_type: Union[Type[str], Type[api.Message], Type[api.AgentQuery]],
        allow_generator: bool,
    ) -> AgentFunc:
        """Constructs an AgentFunc from a Python function.

        Args:
            func: the Python function to wrap
            oauth_params: optional OAuthParams, required if the func requires an OAuthHandler
            default_message_type: the message type to default to if type annotations aren't present
            allow_generator: whether the function is allowed to return multiple responses
        """

        if not inspect.isfunction(func):
            raise TypeError(
                f"Registered function {func!r} is not a function, but a {type(func)!r}."
            )
        signature = inspect.signature(func)
        func_name = func.__name__
        params = signature.parameters

        # Validate that there are no var args (*args or **kwargs).
        if any(
            param.kind in (param.VAR_KEYWORD, param.VAR_POSITIONAL)
            for param in params.values()
        ):
            raise TypeError(
                f"Registered function {func_name} cannot accept variable args: {params!r}."
            )

        # Using the signature (parameter names and type annotations) of the provided function,
        # figure out which parameters correspond to which of the possible values (message,
        # user storage, OAuth handler).
        #
        # Because the types of each of the possible values are distinct, we prioritize type
        # annotations if present, but fall back to well-known names.
        #
        # For each of the possible values we'll try to construct a `BoundArgumentMapper`
        # that can be used at runtime to set the arguments.

        # Ensure every parameter is used exactly once.
        unbound_parameter_names = set(params.keys())

        # Try to bind a query/message mapper.
        typed_message_mappers = {
            api.AgentQuery: lambda query, agent_id: query,
            api.Message: lambda query, agent_id: query.message,
            str: lambda query, agent_id: query.message.text,
        }
        bound_message_mapper = _bind_argument_mapper(
            func,
            unbound_parameter_names=unbound_parameter_names,
            mappers_by_type_or_name=(
                *typed_message_mappers.items(),
                ("query", typed_message_mappers[default_message_type]),
                ("message", typed_message_mappers[default_message_type]),
            ),
        )

        # Try to bind a UserStorage mapper.
        user_storage_mapper = lambda query, agent_id: user_storage.UserStorage(
            query, agent_id
        )
        bound_user_storage_mapper = _bind_argument_mapper(
            func,
            unbound_parameter_names=unbound_parameter_names,
            mappers_by_type_or_name=(
                (user_storage.UserStorage, user_storage_mapper),
                ("user_storage", user_storage_mapper),
            ),
        )

        # Try to bind an OAuthHandler mapper.
        oauth_mapper = lambda query, agent_id: oauth_params and oauth.OAuthHandler(
            oauth_params, query, agent_id
        )
        bound_oauth_mapper = _bind_argument_mapper(
            func,
            unbound_parameter_names=unbound_parameter_names,
            mappers_by_type_or_name=(
                (oauth.OAuthHandler, oauth_mapper),
                ("oauth_handler", oauth_mapper),
            ),
        )
        if bound_oauth_mapper is not None and oauth_params is None:
            raise TypeError(
                f"Function {func_name} that accepts 'oauth_handler' as an argument cannot "
                f"be registered with an agent that hasn't set 'oauth_params' in its "
                "constructor."
            )

        # If we didn't get a message handler and there's a single untyped argument
        # remaining and it's the first parameter, treat it as a message.
        #
        # def f(x): allowed, x is a message
        # def f(x, user_storage): allowed, x is a message
        # def f(user_storage, x): not allowed
        # def f(x, y): not allowed
        # def f(x: int): not allowed
        type_hints = get_type_hints(func)
        if bound_message_mapper is None and len(unbound_parameter_names) == 1:
            remaining_parameter_name = next(iter(unbound_parameter_names))
            first_parameter_name = next(iter(params.keys()))
            if (
                first_parameter_name == remaining_parameter_name
                and remaining_parameter_name not in type_hints
            ):
                bound_message_mapper = (
                    remaining_parameter_name,
                    typed_message_mappers[default_message_type],
                )
                unbound_parameter_names.remove(remaining_parameter_name)

        # All parameters should be accounted for.
        if len(unbound_parameter_names) > 0:
            raise TypeError(
                f"Registered function {func_name} had unknown parameters: {unbound_parameter_names}"
            )

        # Validate that the return type annotation is compatible.
        return_type = get_type_hints(func).get("return")
        if return_type is not None:
            VALID_RETURN_TYPES = (api.AgentResponse, api.Message, str)
            is_return_type_valid = return_type in VALID_RETURN_TYPES

            if allow_generator and not is_return_type_valid:
                # If allow_generator is set, the function can return any Iterable
                # of an allowed type.
                origin_type = typing.get_origin(return_type)
                if origin_type is not None and issubclass(
                    origin_type, collections.abc.Iterable
                ):
                    args = typing.get_args(return_type)
                    is_return_type_valid = (args == ()) or args[0] in VALID_RETURN_TYPES

            if not is_return_type_valid:
                raise TypeError(
                    f"Expected registered function to return an AgentResponse, a Message, "
                    f"or str but it returns {return_type}."
                )

        return AgentFunc(
            func,
            [
                bound_mapper
                for bound_mapper in (
                    bound_message_mapper,
                    bound_user_storage_mapper,
                    bound_oauth_mapper,
                )
                if bound_mapper is not None
            ],
            allow_generator,
        )


def _adapt_func_result(
    result: Any, allow_multiple: bool
) -> Iterable[api.AgentResponse]:
    """Adapts any allowed func return type into an Iterable of AgentResponses."""

    if isinstance(result, api.AgentResponse):
        yield result
    elif isinstance(result, api.Message):
        yield api.AgentResponse(message=result)
    elif isinstance(result, str):
        yield api.AgentResponse(message=api.Message(result))
    elif isinstance(result, collections.abc.Iterable):
        for value in result:
            yield from _adapt_func_result(value, allow_multiple)

            if not allow_multiple:
                break
    else:
        raise TypeError(
            f"The func result was type {type(result)}, but must be a str, Message, AgentResponse, or iterable."
        )


def _bind_argument_mapper(
    func: Callable,
    unbound_parameter_names: Set[str],
    mappers_by_type_or_name: Iterable[Tuple[Union[type, str], ArgumentMapper]],
) -> Optional[BoundArgumentMapper]:
    """Attempts to select an argument name to which a mapper can be bound.

    If a mapper is bound, the name will be removed from `unbound_parameter_names`.
    """

    type_hints = get_type_hints(func)

    for type_or_name, mapper in mappers_by_type_or_name:
        for name in unbound_parameter_names:
            if isinstance(type_or_name, type) and type_hints.get(name) == type_or_name:
                unbound_parameter_names.remove(name)
                return name, mapper
            elif (
                name not in type_hints
                and isinstance(type_or_name, str)
                and name == type_or_name
            ):
                unbound_parameter_names.remove(name)
                return name, mapper

    return None



import os
import pathlib
import random
import string
import subprocess
from unittest import mock

import click.testing
import gql.client
import pytest
import validators

import fixieai
from fixieai import constants
from fixieai.cli import cli
from fixieai.cli.agent import agent_config
from fixieai.client.client import FixieClient

from . import commands


@pytest.fixture(autouse=True)
def mock_fixie_api_key(mocker):
    mocker.patch.object(constants, "fixie_api_key", return_value="TEST")
    os.environ["FIXIE_API_KEY"] = "TEST"


@pytest.fixture(autouse=True)
def mock_graphql(mocker):
    def _execute(document, *args, **kwargs):
        if document.definitions[0].name.value == "getUsername":
            return {"user": {"username": "testuser"}}
        elif document.definitions[0].name.value == "getAgentById":
            return {
                "agentById": {
                    "agentId": "testuser/testagent",
                    "handle": "testagent",
                    "name": "Test Agent",
                    "description": "Test agent description",
                    "queries": [],
                    "funcUrl": "https://example.com",
                    "moreInfoUrl": "https://example.com",
                    "published": False,
                    "owner": {
                        "username": "testuser",
                    },
                }
            }
        elif document.definitions[0].name.value == "UpdateAgent":
            return {
                "updateAgent": {
                    "agent": {
                        "agentId": "testuser/testagent",
                    }
                }
            }
        else:
            raise ValueError(
                "Unmocked GraphQL operation: " + document.definitions[0].name.value
            )

    return mocker.patch.object(gql.client.Client, "execute", side_effect=_execute)


def test_update_agent_requirements_adds_fixie():
    new_requirements = commands._update_agent_requirements([], [])
    assert new_requirements == [commands.CURRENT_FIXIE_REQUIREMENT]


def test_update_agent_requirements_honors_specified_fixie():
    new_requirements = commands._update_agent_requirements([], ["fixieai"])
    assert new_requirements == ["fixieai"]


def test_update_agent_requirements_overrides_existing_fixie():
    new_requirements = commands._update_agent_requirements(
        ["fixieai", "fixieai == 0.0"], []
    )
    assert new_requirements == [commands.CURRENT_FIXIE_REQUIREMENT]


def test_update_agent_requirements_combines():
    new_requirements = commands._update_agent_requirements(
        ["package1", "package2"], ["package2", "package3"]
    )
    assert new_requirements == [
        "package1",
        "package2",
        commands.CURRENT_FIXIE_REQUIREMENT,
        "package3",
    ]


@pytest.mark.parametrize("seed", range(100))
def test_slugify(seed):
    random.seed(seed)

    text = "".join(random.choices(string.ascii_letters, k=40))
    slugified = commands._slugify(text)
    assert len(slugified) <= len(text)
    assert validators.slug(slugified)


def test_init():
    runner = click.testing.CliRunner()
    with runner.isolated_filesystem() as temp_dir:
        result = runner.invoke(commands.init_agent)
        assert result.exit_code == 0

        config = agent_config.load_config(temp_dir)
        assert config.handle == commands._slugify(os.path.basename(temp_dir))
        assert os.path.exists(os.path.join(temp_dir, "main.py"))
        assert os.path.exists(os.path.join(temp_dir, "requirements.txt"))


def test_init_with_path():
    runner = click.testing.CliRunner()
    with runner.isolated_filesystem() as temp_dir:
        result = runner.invoke(commands.init_agent, ["foo/bar/baz"])
        assert result.exit_code == 0

        agent_dir = os.path.join(temp_dir, "foo/bar/baz")
        config = agent_config.load_config(agent_dir)
        assert config.handle == "baz"
        assert os.path.exists(os.path.join(agent_dir, "main.py"))
        assert os.path.exists(os.path.join(agent_dir, "requirements.txt"))


def test_init_args():
    runner = click.testing.CliRunner()
    with runner.isolated_filesystem() as temp_dir:
        result = runner.invoke(
            commands.init_agent,
            [
                "foo",
                "--handle",
                "my-handle",
                "--description",
                "a description",
                "--entry-point",
                "main:entrypoint",
                "--more-info-url",
                "https://example.com",
            ],
        )
        assert result.exit_code == 0
        agent_dir = os.path.join(temp_dir, "foo")
        config = agent_config.load_config(agent_dir)
        assert config.handle == "my-handle"
        assert config.description == "a description"
        assert config.entry_point == "main:entrypoint"
        assert config.more_info_url == "https://example.com"


def test_init_validation_prompting():
    runner = click.testing.CliRunner()
    with runner.isolated_filesystem() as temp_dir:
        result = runner.invoke(
            commands.init_agent,
            ["foo"],
            input="""my invalid handle
my-handle
a description
invalid entry point
main:entrypoint
invalid URL

""",
        )
        assert result.exit_code == 0
        agent_dir = os.path.join(temp_dir, "foo")
        config = agent_config.load_config(agent_dir)
        assert config.handle == "my-handle"
        assert config.description == "a description"
        assert config.entry_point == "main:entrypoint"
        assert config.more_info_url == ""


@pytest.mark.parametrize(
    "command",
    [
        ["agent", "serve", "--no-tunnel"],
        ["agent", "serve", "foo/bar/baz", "--no-tunnel"],
        ["agent", "deploy", "--no-validate"],
        ["agent", "deploy", "foo/bar/baz", "--no-validate"],
        # venv provisioning takes a long time, so only test it once per command.
        ["agent", "serve", "foo/bar/baz", "--no-tunnel", "--venv"],
        ["agent", "deploy", "foo/bar/baz", "--validate"],
    ],
)
def test_implicit_init(mocker, command):
    runner = click.testing.CliRunner()
    with runner.isolated_filesystem():
        subprocess_run = subprocess.run

        def mock_subprocess(cmd_args, *args, **kwargs):
            if "uvicorn" in cmd_args:
                # Mock out the uvicorn process.
                return mock.MagicMock()

            return subprocess_run(cmd_args, *args, **kwargs)

        mocker.patch.object(subprocess, "run", side_effect=mock_subprocess)

        # Use the current version of fixieai as the fixie dependency.
        mocker.patch.object(
            commands,
            "CURRENT_FIXIE_REQUIREMENT",
            str(pathlib.Path(fixieai.__file__).parent.parent),
        )

        # Mock out deployment API calls.
        mocker.patch.object(FixieClient, "deploy_agent", return_value=None)
        mocker.patch.object(FixieClient, "refresh_agent", return_value=None)

        result = runner.invoke(
            cli.fixie,
            command,
        )
        assert result.exit_code == 0
